import sys, logging
from pyuia import PageObject, cacheable
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.common.by import By

__all__ = ['SeleniumPageObject', 'find_by', 'cacheable']
_logger = logging.getLogger(__name__)

class SeleniumPageObject(PageObject):

    def __init__(self, context):
        PageObject.__init__(
            self, context,
            (NoSuchElementException, StaleElementReferenceException))

    @property
    def _driver(self):
        return self._context.driver

    def _is_displayed(self, element):
        try:
            displayed = element.is_displayed()
            _logger.debug('Element (%s) is displayed? %s.', element.id, displayed)
            return displayed
        except StaleElementReferenceException:
            logging.debug('Element (%s) is NOT displayed because of stale reference.', element.id)
            return False

_strategy_kwargs = {
    'id_': By.ID,
    'xpath': By.XPATH,
    'link_text': By.LINK_TEXT,
    'partial_link_text': By.PARTIAL_LINK_TEXT,
    'name': By.NAME,
    'tag_name': By.TAG_NAME,
    'class_name': By.CLASS_NAME,
    'css_selector': By.CSS_SELECTOR }

from pyuia import cacheable as cacheable_decorator # naming conflict between global and parameter names

def find_by(how=None, using=None, multiple=False, cacheable=True, if_exists=False,
            context=None, driver_attr='_driver', **kwargs):
    """Create a callable which can be evaluated lazily to find UI elements.

    This function implements the concept mentioned in Page Factory (or PageFactory) pattern (https://code.google.com/p/selenium/wiki/PageFactory). It helps to reduce the amount of boilerplate code while implementing page objects. For more details, see https://jeremykao.wordpress.com/2015/06/10/pagefactory-pattern-in-python/.

    Args:
        how: Locator strategy. It can be one of the class variables defined in
            `selenium.webdriver.common.by.By`, e.g., `By.ID`, `By.NAME` etc.
        using: The locator.
        multiple: Whether the lookup return multiple elements.
        cacheable: Whether to cache the result.
        if_exists: Whether to return `None` to indicate the element doesn't exist,
            instead of raising an exception.
        context: The starting point of a search.
        driver_attr: The attribute name for getting the reference to WebDriver. Defaults to '_driver'.

    Kwargs:
        The following keyword arguments are supported for various locator strategies: id_ (to avoid conflict with the built-in keyword id), name, class_name, css_selector, tag_name, xpath, link_text, and partial_link_text.

    Returns: A callable which can be evaluated lazily to find UI elements.

    """
    # 'how' AND 'using' take precedence over keyword arguments
    _how, _using = how, using
    if not (_how and _using):
        if len(kwargs) != 1 or kwargs.keys()[0] not in _strategy_kwargs.keys() :
            raise ValueError(
                "If 'how' AND 'using' are not specified, one and only one of the following "
                "valid keyword arguments should be provided: %s." % _strategy_kwargs.keys())

        key = kwargs.keys()[0]
        _how, _using = _strategy_kwargs[key], kwargs[key]

    def func(page_object):
        driver = getattr(page_object, driver_attr)

        # ctx - driver or a certain element
        if context is None:
            ctx = driver
            container = None
        elif callable(context):
            container = ctx = context(page_object)
            if not ctx:
                if if_exists:
                    return None
                else:
                    raise NoSuchElementException("The element as the context doesn't exist.")
        else: # element
            container = ctx = context

        lookup = ctx.find_elements if multiple else ctx.find_element

        try:
            return lookup(_how, _using)
        except NoSuchElementException as e:
            if if_exists: return None
            msg = "%s ; find_by(how='%s', using='%s', multiple=%s, cacheable=%s, " \
                  "if_exists=%s, context=%s)" % \
                  (str(e), _how, repr(_using), multiple, cacheable, if_exists, context)
            raise NoSuchElementException(msg), None, sys.exc_info()[2]

    func = cacheable_decorator(func, cache_none=not if_exists) if cacheable else func

    # for debugging, expose criteria of the lookup
    func.__name__ = "find_by(how='%s', using=%s, multiple=%s, cacheable=%s, " \
                    "if_exists=%s, context=%s)" % \
                    (_how, repr(_using), multiple, cacheable, if_exists, context)
    return func




import logging, inspect
from robot.utils import ConnectionCache
from pyuia import PageObject, get_page_object
from util import is_test_failed, log_screenshot, log_text, in_context as in_robot_context

__all__ = ['BaseAppLibrary']
_logger = logging.getLogger(__name__)

def _state_capturing_decorator(method):
    def decorator(*args, **kwargs):
        name = method.__name__
        # accessor = any(name.startswith(prefix) or name == prefix for prefix in ['is_', 'get_', 'should_be_'])
        accessor = False
        self = args[0] # the keyword library itself

        self._capture_state()
        try:
            result = method(*args, **kwargs)

            # if the the result is a page object, update the current page.
            if isinstance(result, PageObject):
                self._current_page = result
        except Exception as err:
            self._capture_state(after=True, err=err)
            raise

        if not accessor: # efficiency
            self._capture_state(after=True)
        return result

    return decorator

class _StateCapturing(type):

      def __new__(cls, clsname, bases, attrs):
          white_list = [
              'open_session',
              'close_session',
              'close_all_sessions',
              'switch_device',
              'open_app',
              'close_app',
          ]

          for name, obj in attrs.items():
              if not (inspect.isroutine(obj) and not name.startswith('_')): continue
              if name in white_list: continue
              attrs[name] = _state_capturing_decorator(obj)
          return type.__new__(cls, clsname, bases, attrs)

class BaseAppLibrary(object):

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'

    if in_robot_context:
        __metaclass__ = _StateCapturing

    def __init__(self):
        self._cache = ConnectionCache()

    def open_session(self, device_id, alias=None):
        """Open a session.

        ``device_id`` is an identifier for looking up configurations of a specific device.

        The optional ``alias`` provided here can be used to switch between sessions/devices later.
        See `Switch Device` for more details.

        """
        _logger.info('Open session; device ID = [%s], alias = [%s])', device_id, alias)
        # init context and install delegates
        context = self._init_context(device_id)
        context._log_screenshot_delegate = self._log_screenshot_delegate
        context._log_page_source_delegate = self._log_page_source_delegate
        self._cache.register(RFConnectionCache(context), alias)

    def open_app(self, reset=None):
        """Open the app.

        To reset app state prior to opening the app, pass a non-empty string to ``reset`` argument.

        Examples:

        | Open App | reset | # reset app state        |
        | Open App |       | # do not reset app state |

        """
        msg = 'App logs (initial)'
        context = self._current_context
        context.open_app(bool(reset))

        context.logs_all = [] # accumulate logs of each step
        log_text('\n'.join(context.get_initial_logs()), msg, 'app_logs_initial_', '.log', level=logging.INFO)

    def _init_context(self):
        raise NotImplementedError()

    def _log_screenshot_delegate(self, msg, *args, **kwargs):
        level = kwargs['level'] if 'level' in kwargs else logging.DEBUG
        if not _logger.isEnabledFor(level):
            return

        page = kwargs['page'] if 'page' in kwargs else None
        msg = msg % args

        if page: msg += ' (%s)' % page.__class__.__name__
        log_screenshot(self._current_context.take_screenshot_as_png(), msg, level=level)

    def _log_page_source_delegate(self, msg, *args, **kwargs):
        level = kwargs['level'] if 'level' in kwargs else logging.DEBUG
        if not _logger.isEnabledFor(level):
            return

        page = kwargs['page'] if 'page' in kwargs else None
        msg = msg % args

        if page: msg += ' (%s)' % page.__class__.__name__
        source, ext = self._current_context.dump_page_source()
        log_text(source, msg, prefix='page_source_', suffix='.%s' % ext, level=level)

    def close_session(self):
        """Terminate current session."""
        self._cache.current.close()

    def close_all_sessions(self):
        """Terminate all open sessions."""
        self._cache.close_all()

    def close_app(self):
        """Close the app."""
        self._cache.current.close_app()

    def switch_device(self, alias):
        """Switch between sessions/devices using alias.

        Examples:

        | Open App      | A | # current session/device is A      |
        | Open App      | B | # current session/device becomes B |
        | Switch Device | A | # switch back to A                 |

        """
        self._cache.switch(alias)

    def _capture_state(self, after=False, err=None):
        # To increase efficiency, screenshots are no longer taken automatically.
        # Developers should explicitly do that AFTER the UI has been changed.
        if not after: return

        failed = bool(err)
        try:
            context = self._current_context
            msg = 'App logs (step, keyword failed? %s)' % failed

            logs_step = context.get_new_logs()
            context.logs_all.extend(logs_step)
            log_text('\n'.join(logs_step), msg, 'app_logs_step_', '.log', level=logging.INFO)
            if not failed: return

            context.log_page_source('Page source', level=logging.INFO)
            context.log_screenshot('Screenshot', level=logging.INFO)
        except:
            _logger.warning('Fail to capture state. (keyword failed = %s)', failed, exc_info=True)

    @property
    def _current_context(self):
        return self._cache.current._context

    @property
    def _current_page(self):
        return self._cache.current._context.current_page

    @_current_page.setter
    def _current_page(self, page):
        self._cache.current._context.current_page = page

class RFConnectionCache(object):

    def __init__(self, context):
        self._context = context

    def close(self):
        self._context.quit()

    def close_app(self):
        # all statements suppress possible errors, or other sessions won't be closed.
        self._capture_state()
        self._context.close_app()

    def _capture_state(self):
        failed = None

        try:
            failed = is_test_failed()
            context = self._context
            msg = 'App logs (about to quit, test failed? %s)' % failed

            context.logs_all.extend(context.get_new_logs())
            log_text('\n'.join(context.logs_all), msg, 'app_logs_all_', '.log', level=logging.INFO)
            if not failed: return

            context.log_page_source('Page source (test failed)', level=logging.INFO)
            context.log_screenshot('Screenshot (test failed)', level=logging.INFO)
        except Exception as e:
            _logger.warning('Fail to capture state. (test failed = %s)', failed, exc_info=True)




import time
from ..selenium import SeleniumPageObject, cacheable
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.common.by import By

__all__ = ['AppiumPageObject', 'find_by', 'cacheable']

class AppiumPageObject(SeleniumPageObject):

    def _press_menu(self):
        self._driver.keyevent(82)
        time.sleep(1) # or then menu may not in the subsequent screenshot.

    def _press_back(self):
        self._driver.keyevent(4)

_strategy_kwargs = {
    'id_': By.ID,
    'xpath': By.XPATH,
    'link_text': By.LINK_TEXT,
    'partial_link_text': By.PARTIAL_LINK_TEXT,
    'name': By.NAME,
    'tag_name': By.TAG_NAME,
    'class_name': By.CLASS_NAME,
    'css_selector': By.CSS_SELECTOR }

from pyuia import cacheable as cacheable_decorator # naming conflict between global and parameter names

def find_by(how=None, using=None, multiple=False, cacheable=True, if_exists=False,
            context=None, scrollable=False, scroll_forward=True, scroll_vertically=True,
            scroll_starting_padding=None, scroll_ending_padding=None, maximum_scrolls=5, driver_attr='_driver', **kwargs):
    """Create a callable which can be evaluated lazily to find UI elements.

    This function implements the concept mentioned in Page Factory (or PageFactory) pattern (https://code.google.com/p/selenium/wiki/PageFactory). It helps to reduce the amount of boilerplate code while implementing page objects. For more details, see https://jeremykao.wordpress.com/2015/06/10/pagefactory-pattern-in-python/.

    Args:
        how: Locator strategy. It can be one of the class variables defined in
            `selenium.webdriver.common.by.By`, e.g., `By.ID`, `By.NAME` etc.
        using: The locator.
        multiple: Whether the lookup return multiple elements.
        cacheable: Whether to cache the result.
        if_exists: Whether to return `None` to indicate the element doesn't exist,
            instead of raising an exception.
        context: The starting point of a search, and the scrollable container
            if argument `scrollable` is set to `True` explicitly.
        scrollable: Whether to perform scroll actions if the element is not found.
            Defaults to `False`.
        scroll_forward: Whether to scroll forward or backward. Defaults to `True`.
        scroll_vertically: Whether to scroll vertically or horizontally. Defaults to `True`.
        scroll_starting_padding: No-touch starting zone. Defaults to `None`.
        scroll_ending_padding: No-touch ending zone. Defaults to `None`.
        maximum_scrolls: The maximum number of attempts to scroll. Defaults to 5.
        driver_attr: The attribute name for getting the reference to WebDriver. Defaults to '_driver'.

    Kwargs:
        The following keyword arguments are supported for various locator strategies: id_ (to avoid conflict with the built-in keyword id), name, class_name, css_selector, tag_name, xpath, link_text, and partial_link_text.

    Returns: A callable which can be evaluated lazily to find UI elements.

    """
    # 'how' AND 'using' take precedence over keyword arguments

    # If _how == 'name', we will replace it by 'xpath' in func.
    # However, since variable cannot be updated in different scope (i.e., find_by & func here),
    # We use dictionary instead.
    method_dic = {'_how':how, '_using': using}

    if not (method_dic['_how'] and method_dic['_using']):
        if len(kwargs) != 1 or kwargs.keys()[0] not in _strategy_kwargs.keys() :
            raise ValueError(
                "If 'how' AND 'using' are not specified, one and only one of the following "
                "valid keyword arguments should be provided: %s." % _strategy_kwargs.keys())

        key = kwargs.keys()[0]
        method_dic['_how'], method_dic['_using'] = _strategy_kwargs[key], kwargs[key]

    def func(page_object):
        driver = getattr(page_object, driver_attr)

        # For appium v1.5.0+, since it doesn't support find by name strategy, we have to adjust our pyuia
        # We replace name with xpath
        # Github release note: https://github.com/appium/appium/releases/tag/v1.5.0
        # Discuss thread: https://discuss.appium.io/t/appium-1-5-fails-to-find-element-by-name/8857/10

        if method_dic['_how'] == 'name':
            method_dic['_how'] = 'xpath'
            method_dic['_using'] = "//*[@text='" + method_dic['_using'] + "' or @content-desc='" + method_dic['_using'] + "']"

        # ctx - driver or a certain element
        if context is None:
            ctx = driver
            container = None
        elif callable(context):
            container = ctx = context(page_object)
            if not ctx:
                if if_exists:
                    return None
                else:
                    raise NoSuchElementException("The element as the context doesn't exist.")
        else: # element
            container = ctx = context

        lookup = ctx.find_elements if multiple else ctx.find_element

        scrolls = 0;

        while True:
            try:
                return lookup(method_dic['_how'], method_dic['_using'])

            except NoSuchElementException as e:

                if not scrollable or scrolls == maximum_scrolls:
                    if if_exists: return None
                    msg = "%s ; find_by(how='%s', using='%s', multiple=%s, cacheable=%s, " \
                          "if_exists=%s, context=%s, scrollable=%s)" % \
                          (str(e), method_dic['_how'], method_dic['_using'], multiple, cacheable, if_exists, context, scrollable)
                    raise NoSuchElementException(msg), None, sys.exc_info()[2]

                scroller = _get_scroller(page_object, container, scrollable)
                _scroll(page_object, driver, scroller, scroll_forward,
                        scroll_vertically, scroll_starting_padding, scroll_ending_padding)
                scrolls += 1

    func = cacheable_decorator(func, cache_none=not if_exists) if cacheable else func

    # for debugging, expose criteria of the lookup
    func.__name__ = "find_by(how='%s', using=%s, multiple=%s, cacheable=%s, " \
                    "if_exists=%s, context=%s, scrollable=%s)" % \
                    (method_dic['_how'], repr(method_dic['_using']), multiple, cacheable, if_exists, context, scrollable)
    return func

def _get_scroller(page_object, container, scrollable):
    if callable(scrollable): # find_by
        scroller = scrollable(page_object)
        if not scroller:
            raise ValueError("The element as the scrollable container doesn't exist.")
        return scroller
    elif scrollable is True:
        if container is None:
            raise ValueError("The argument 'context' is mandatory if argument 'scrollable' is set to True explicitly.")
        else:
            return container
    else: # element
        return scrollable

def _scroll(page_object, driver, scroller, forward, vertically, starting_padding, ending_padding):
    loc, size = scroller.location, scroller.size
    x, y, w, h = loc['x'], loc['y'], size['width'], size['height']

    if vertically:
        points = [(x + w / 2, y + h - 1), (x + w / 2, y + 1)]
    else:
        points = [(x + w - 1, y + h / 2), (x + 1, y + h / 2)]

    if not forward:
        points.reverse()

    x1, y1 = points[0]
    x2, y2 = points[1]

    # take margin (no-touch zone) into account
    if isinstance(starting_padding, (int, float)):
        if isinstance(starting_padding, float):
            starting_padding = int((h if vertically else w) * starting_padding)

        offset = -starting_padding if forward else starting_padding

        if vertically:
            y1 += offset
        else:
            x1 += offset
    elif starting_padding is not None:
        starting_padding = starting_padding(page_object) if callable(starting_padding) else starting_padding
        if starting_padding is not None:
            loc, size = starting_padding.location, starting_padding.size
            x, y, w, h = loc['x'], loc['y'], size['width'], size['height']

            if vertically:
                y1 = y - 1 if forward else y + h + 1
            else:
                x1 = x - 1 if forward else x + w + 1

    if isinstance(ending_padding, (int, float)):
        if isinstance(ending_padding, float):
            ending_padding = int((h if vertically else w) * ending_padding)

        offset = ending_padding if forward else -ending_padding

        if vertically:
            y2 += offset
        else:
            x2 += offset
    elif ending_padding is not None:
        ending_padding = ending_padding(page_object) if callable(ending_padding) else ending_padding
        if ending_padding is not None:
            loc, size = ending_padding.location, ending_padding.size
            x, y, w, h = loc['x'], loc['y'], size['width'], size['height']

            if vertically:
                y2 = y + h + 1 if forward else y - 1
            else:
                x2 = x + w + 1 if forward else x - 1
    driver.swipe(x1, y1, x2, y2, abs(x1-x2+y1-y2)*3)



from ..selenium import SeleniumContext
from .util import get_logs

__all__ = ['AppiumContext']

class AppiumContext(SeleniumContext):

    def dump_page_source(self):
        return (self.driver.page_source, 'xml')

    def get_new_logs(self):
        log_type = 'syslog' if self.platform == 'iOS' else 'logcat'
        return get_logs(self.driver, log_type)

    def open_app(self, reset):
        if reset:
            # "remove -> launch" is more efficient than "launch -> reset"
            self._remove_app()
            self.driver.launch_app()
        else:
            self.driver.launch_app()

    def _remove_app(self):
        caps = self.driver.capabilities
        platform = caps['platformName']
        if platform == 'Android':
            self.driver.remove_app(caps['appPackage'])
        elif platform == 'iOS':
            self.driver.remove_app(caps['bundleId'])
        else:
            assert False, platform

    def close_app(self):
        self.driver.close_app()




import time, logging
from .exceptions import TimeoutError, ElementNotFoundError

__all__ = ['PageObject', 'get_page_object', 'cacheable']
_logger = logging.getLogger(__name__)

_page_singletons = {} # cache

def get_page_object(page_class, context):
    fqcn = '%s.%s' % (page_class.__module__, page_class.__name__)
    _logger.debug('Get page object; FQCN = %s', fqcn)
    if fqcn in _page_singletons:
        cache = _page_singletons[fqcn]
        if cache._context is context:
            _logger.debug('Cached, and the context remains unchanged.')
            cache._invalidate_elements_cache()

            return cache
        else:
            _logger.debug('Cached, but the context is invalid.')

    page = page_class(context)
    _page_singletons[fqcn] = page
    return page

def _is_iterable(obj):
    try:
       iter(obj)
       return True
    except TypeError:
       return False

def _to_iterable(obj):
    return obj if _is_iterable(obj) else (obj,)

_NOT_FOUND_EXCEPTIONS = (ElementNotFoundError,)

_ELEMENTS_CACHE_ATTR = '_pyuia_elements_cache'

def cacheable(lookup, cache_none=True):
    def func(self):
        if not hasattr(self, _ELEMENTS_CACHE_ATTR):
            setattr(self, _ELEMENTS_CACHE_ATTR, {}) # {callable_id: element(s)}
        cache = getattr(self, _ELEMENTS_CACHE_ATTR)

        key = id(lookup)
        if key not in cache:
            result = lookup(self)
            if result is None and not cache_none: return
            cache[key] = result
        return cache[key]

    return func

class PageObject(object):

    _WAIT_INTERVAL = 0
    _WARN_TIMEOUT = 5
    _WAIT_TIMEOUT = 10
    _PAGE_WARN_TIMEOUT = 5
    _PAGE_WAIT_TIMEOUT = 10
    
    def __init__(self, context, not_found_exceptions):
        self._context = context

        exceptions = list(_NOT_FOUND_EXCEPTIONS)
        if not_found_exceptions:
            try:
                exceptions.extend(not_found_exceptions)
            except TypeError: # not iterable
                exceptions.append(not_found_exceptions)
        self._not_found_exceptions = tuple(exceptions)

        exceptions.append(AssertionError)
        self._page_assertion_exceptions = tuple(exceptions)

    def _go_to(self, page_class):
        """Instantiate a page object."""
        page = get_page_object(page_class, self._context)
        page._from_page_class = self.__class__

        page.wait_for_page_loaded(self.__class__)
        return page

    def _back_to(self, page_class=None):
        if not page_class:
            if not hasattr(self, '_from_page_class'):
                raise RuntimeError("_back_to(page_class) don't know where to go. You can explicitly specify "
                                   "'page_class' or implement page transition with _go_to(page_class).")
            page_class = self._from_page_class
        page = get_page_object(page_class, self._context)

        page.wait_for_page_loaded(self.__class__)
        return page

    def _invalidate_elements_cache(self):
        if hasattr(self, _ELEMENTS_CACHE_ATTR):
            delattr(self, _ELEMENTS_CACHE_ATTR)

    def wait_for_page_loaded(self, from_page_class=None, timeout_warn=None, timeout=None):
        timeout_warn = timeout_warn or self._PAGE_WARN_TIMEOUT
        timeout = timeout or self._PAGE_WAIT_TIMEOUT

        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        handlers = self._get_page_entry_handlers(from_page_class)

        warned = False
        while True:
            try:
                self._invalidate_elements_cache()
                self.assert_on_this_page(from_page_class)
                break
            except self._page_assertion_exceptions:
                if not warned and time.time() > timeout_warn:
                    self._log_screenshot(
                        'Wait for page loaded. Time elapsed = [%s]s.',
                        time.time() - start_time,
                        level=logging.WARN)
                    warned = True

                handlers = self._consult_handlers(handlers)
                time.sleep(self._WAIT_INTERVAL)
                if time.time() > timeout: raise

        self._log_screenshot('Already on the page.')

        # return True to indicate UI changed.
        if self._on_page_entry(from_page_class):
            self._log_screenshot('Page loaded.')

        return self

    def _get_page_entry_handlers(self, from_page_class): pass

    def assert_on_this_page(self, from_page_class): pass

    def _on_page_entry(self, from_page_class):
        """To put the page in a known state."""
        pass

    def _log_screenshot(self, msg, *args, **kwargs):
        kwargs['page'] = self
        self._context.log_screenshot(msg, *args, **kwargs)

    def _log_page_source(self, msg, *args, **kwargs):
        kwargs['page'] = self
        self._context.log_page_source(msg, *args, **kwargs)

    def _assert_present(self, locators, check_visibility=False):
        single_loc = not _is_iterable(locators)
        locators = _to_iterable(locators)
        elements = []

        for locator in locators:
            try:
                element = locator()
            except self._not_found_exceptions as e:
                _logger.debug(
                    'Assert ALL present. The locator (%s) did not resolve to '
                    'an element.', locator)
                element = None
            if not element:
                assert False, locator # None or empty sequence

            if check_visibility and not self._is_displayed(element):
                assert False, locator
            elements.append(element)

        return elements[0] if single_loc else elements

    def _assert_visible(self, locators):
        return self._assert_present(locators, check_visibility=True)

    def _assert_any_present(self, locators, check_visibility=False):
        locators = _to_iterable(locators)

        for locator in locators:
            try:
                element = locator()
            except self._not_found_exceptions as e:
                _logger.debug(
                    'Assert ANY present. The locator (%s) did not resolve to '
                    'an element.', locator)
                element = None
            if not element: continue # None or empty sequence

            if check_visibility and not self._is_displayed(element): continue
            return element

        assert False, locators

    def _assert_any_visible(self, locators):
        return self._assert_any_present(locators, check_visibility=True)

    def _wait_present(self, locators, timeout_warn=None, handlers=None,
                      timeout=None, check_visibility=False):
        timeout_warn = timeout_warn or self._WARN_TIMEOUT
        timeout = timeout or self._WAIT_TIMEOUT

        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        single_loc = not _is_iterable(locators)
        locators = _to_iterable(locators)

        warned = False
        while True:
            elements = []
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ALL present. The locator (%s) did not resolve to '
                        'an element.', locator)
                    element = None
                if not element: break # None or empty sequence

                if check_visibility and not self._is_displayed(element): break
                elements.append(element)

            if len(elements) == len(locators):
                return elements[0] if single_loc else elements

            if not warned and time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ALL elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.',
                    locators, check_visibility, time.time() - start_time,
                    level=logging.WARN)
                warned = True
            handlers = self._consult_handlers(handlers)

            time.sleep(self._WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ALL elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility, time.time() - start_time))

    def _wait_visible(self, locators, timeout_warn=None, handlers=None, timeout=None):
        return self._wait_present(locators, timeout_warn, handlers, timeout, check_visibility=True)

    def _wait_any_present(self, locators, timeout_warn=None, handlers=None,
                          timeout=None, check_visibility=False):
        timeout_warn = timeout_warn or self._WARN_TIMEOUT
        timeout = timeout or self._WAIT_TIMEOUT

        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        locators = _to_iterable(locators)

        warned = False
        while True:
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ANY present. The locator (%s) did not resolve to '
                        'an element.', locator)
                    element = None
                if not element: continue # None or empty sequence

                if check_visibility and not self._is_displayed(element): continue
                return element

            if not warned and time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ANY present. locators = %s, time elapsed = [%s]s.',
                    locators, time.time() - start_time, level=logging.WARN)
                warned = True
            handlers = self._consult_handlers(handlers)

            time.sleep(self._WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ANY elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility, time.time() - start_time))

    def _wait_any_visible(self, locators, timeout_warn=None, handlers=None,
                          timeout=None):
        return self._wait_any_present(locators, timeout_warn, handlers, timeout, check_visibility=True)

    def _wait_absent(self, locators, timeout_warn=None, minwait=3,
                     handlers=None, timeout=None, check_visibility_only=False):
        timeout_warn = timeout_warn or self._WARN_TIMEOUT
        timeout = timeout or self._WAIT_TIMEOUT

        start_time = time.time()
        timeout_appear = start_time + minwait
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        locators = _to_iterable(locators)

        warned = False
        while True:
            # to avoid the situation that elements are absent simply because
            # other elements such as error dialogs are displayed.
            handlers = self._consult_handlers(handlers)
            any_invalid = False
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ALL absent. The locator (%s) did not resolve to '
                        'an element.', locator)
                    element = None
                if not element: continue

                if not check_visibility_only or self._is_displayed(element):
                    any_invalid = True
                    break

            # wait for at least 'minwait' seconds to make sure target
            # element(s) won't appear at this time.
            if not any_invalid and time.time() > timeout_appear: return
            if not warned and time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ALL elements to be absent. locators = %s, '
                    'check_visibility_only = [%s], time elapsed = [%s]s.',
                    locators, check_visibility_only, time.time() - start_time, 
                    level=logging.WARN)
                warned = True

            time.sleep(self._WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ALL elements to be absent. locators = %s, '
                    'check_visibility_only = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility_only, time.time() - start_time))

    def _wait_invisible(self, locators, timeout_warn=None, minwait=3,
                        handlers=None, timeout=None):
        self._wait_absent(locators, timeout_warn, minwait, handlers, timeout, check_visibility_only=True)

    def _watch(self, handlers, max_duration=5):
        timeout = time.time() + max_duration
        while True:
            handlers = self._consult_handlers(handlers)
            if not handlers: break
            if time.time() > timeout: break

    def _consult_handlers(self, handlers):
        if not handlers: return

        # convert handlers to a mutable list of (locator, handler)
        if isinstance(handlers, dict):
            handlers = handlers.items()
        handlers = list(handlers)
        _logger.debug('Consult handlers. handlers = %s.', [h[0] for h in handlers])

        # consult a handler at a time (rotation)
        locator, handler = handlers[0]

        try:
            element = locator()
        except self._not_found_exceptions as e:
            _logger.debug('The locator (%s) did not resolve to an element.', locator)
            element = None

        # consult the handler again later, or drop it.
        del handlers[0]
        if not element or not self._is_displayed(element) or handler(element):
            handlers.append((locator, handler))

        _logger.debug('Rotated/modified handlers: %s', [h[0] for h in handlers])
        return handlers

